
#include "setup.h"

#define CSTP_VERSION  "1.4"
#define CSTP_DEFAULT_CONNECT_TIMEOUT	90
#define CSTP_DEFAULT_QUEUE_LIMIT		(1024*1024)

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/stat.h>
#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_POLL
#ifdef HAVE_POLL_H
#include <poll.h>
#else
#include <sys/poll.h>
#endif
#else

struct pollfd
{
	int		fd;
	int		events;
	int		revents;
};

#define POLLIN  0x0001
#define POLLOUT 0x0004

#endif
#include <stdarg.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <assert.h>

RCSID("$Id: cstcpproxy.c,v 1.9 2001/10/15 01:26:17 cbehrens Exp $");

typedef struct _tcpproxy_qe			TCPPROXY_QE;
typedef struct _tcpproxy_q			TCPPROXY_Q;
typedef struct _tcpproxy_polltype	TCPPROXY_POLLTYPE;
typedef struct _tcpproxy_conn		TCPPROXY_CONN;
typedef struct _tcpproxy_listener	TCPPROXY_LISTENER;
typedef struct _tcpproxy_allow		TCPPROXY_ALLOW;
typedef struct _tcpproxy_proxy		TCPPROXY_PROXY;
typedef struct _tcpproxy_config		TCPPROXY_CONFIG;

struct _tcpproxy_qe
{
	char			*qe_data;
	int				qe_pos;
	int				qe_len;
	int				qe_size;

	TCPPROXY_QE		*qe_next;
};

struct _tcpproxy_q
{
	int				q_len;
	TCPPROXY_QE		*q_entries;
	TCPPROXY_QE		*q_last_entry;
};

struct _tcpproxy_polltype
{
#define TCPPROXY_PT_TYPE_CONN		0
#define TCPPROXY_PT_TYPE_PROXY		1
	char	pt_type;
	void	*pt_vptr;
};

struct _tcpproxy_conn
{
#define TCPPROXY_TPC_FLAGS_CONNECTING		0x001
	int					tpc_flags;

	time_t				tpc_conntime;

	int					tpc_afd;
	int					tpc_afd_pfds_pos;
	TCPPROXY_Q			tpc_afd_outqueue;

	int					tpc_cfd;
	int					tpc_cfd_pfds_pos;
	TCPPROXY_Q			tpc_cfd_outqueue;

	struct sockaddr_in	tpc_rem_addr;

	TCPPROXY_POLLTYPE	tpc_polltype;

	TCPPROXY_CONFIG		*tpc_config;
	TCPPROXY_PROXY		*tpc_proxy;

	TCPPROXY_CONN		*tpc_prev;
	TCPPROXY_CONN		*tpc_next;
};

struct _tcpproxy_allow
{
	struct in_addr		tpa_inaddr;

	TCPPROXY_ALLOW		*tpa_next;
	TCPPROXY_ALLOW		*tpa_prev;
};

struct _tcpproxy_proxy
{
	int					tpp_lfd;
	int					tpp_lfd_pfds_pos;

	struct sockaddr_in	tpp_src_addr;
	struct sockaddr_in	tpp_dest_addr;
	struct sockaddr_in	tpp_bind_addr;

	TCPPROXY_POLLTYPE	tpp_polltype;

	TCPPROXY_CONFIG		*tpp_config;

	TCPPROXY_CONN		*tpp_conns;

	TCPPROXY_PROXY		*tpp_next;

	TCPPROXY_ALLOW		*tpp_allows;

	int					tpp_queue_limit;
	int					tpp_connect_timeout;
};

struct _tcpproxy_config
{
	int					tpc_flags;

	struct pollfd		*tpc_pfds;
	int					tpc_pfds_num;
	int					tpc_pfds_size;

	time_t				tpc_timeout_check;

	TCPPROXY_POLLTYPE	**tpc_conns;
	int					tpc_conns_size;

	TCPPROXY_PROXY		*tpc_proxies;

	TCPPROXY_ALLOW		*tpc_allows;

	int					tpc_queue_limit;
	int					tpc_connect_timeout;
};

static int	_TP_LogSyslog = 0;
static int	_TP_SigHup = 0;
static int	_TP_SigTerm = 0;

static char *_my_strerror(int errnum);
#ifndef HAVE_POLL
static int poll(struct pollfd *pfd, int nfds, int timeout);
#endif
static void _tp_log(char *format, ...);
static char *_tp_next_arg(char **args);
static void _tp_free_conn(TCPPROXY_CONN *tpconn);
static void _tp_free_conns(TCPPROXY_PROXY *tpp);
static void _tp_free_allows(TCPPROXY_ALLOW **allows);
static int _tp_allowed(TCPPROXY_ALLOW *allows, struct sockaddr_in *sin);
static void _tp_config_free(TCPPROXY_CONFIG **tpc);
static int __tp_config_load(char *filename, TCPPROXY_CONFIG **tpcptr);
static int _tp_config_load(char *filename, TCPPROXY_CONFIG **tpcptr);
static int _tp_queue_fill_iovec(TCPPROXY_Q *queue, struct iovec *iov, int num);
static int _tp_queue_del(TCPPROXY_Q *queue, int amt);
static int _tp_queue_add(TCPPROXY_Q *queue, char *data, int datalen);
static int _tp_add_conn(TCPPROXY_PROXY *tpp, int fd, struct sockaddr_in *sin);
static int _tp_conn_complete(TCPPROXY_CONN *tpc, struct pollfd *pfd);
static void _tp_ioloop(TCPPROXY_CONFIG *tpc);
static void _usage(int l);

static char *_my_strerror(int errnum)
{
#ifdef HAVE_STRERROR
    if (errnum < 0)
        return "Undefined error";
    return strerror(errnum);
#else
    return (errnum <= 0 || errnum > sys_nerr) ? "Undefined error" : sys_errlist[
errnum];
#endif
}

#ifndef HAVE_POLL

static int poll(struct pollfd *pfd, int nfds, int timeout)
{
	struct timeval  tv;
	int				ret;
	int				i;
	fd_set			readset;
	fd_set			writeset;
	int				highfd = -1;

	tv.tv_sec = timeout/1000;
	tv.tv_usec = timeout % 1000;

	FD_ZERO(&readset);
	FD_ZERO(&writeset);

	for(i=0;i<nfds;i++)
	{
		if (pfd[i].events & POLLIN)
		{
			FD_SET(pfd[i].fd, &readset);
			if (pfd[i].fd > highfd)
				highfd = pfd[i].fd;
		}
		if (pfd[i].events & POLLOUT)
		{
			FD_SET(pfd[i].fd, &writeset);
			if (pfd[i].fd > highfd)
				highfd = pfd[i].fd;
		}
	}

	ret = select(highfd + 1, &readset, &writeset, NULL,
					(timeout == -1) ? NULL : &tv);

	if (ret <= 0)
		return ret;

	for(i=0;i<nfds;i++)
	{
		if (FD_ISSET(pfd[i].fd, &readset))
			pfd[i].revents |= POLLIN;
		if (FD_ISSET(pfd[i].fd, &writeset))
			pfd[i].revents |= POLLOUT;
	}
	return ret;
}

#endif

static RETSIGTYPE _sig_hup(int sig)
{
	_TP_SigHup = 1;
	signal(sig, _sig_hup);
}

static RETSIGTYPE _sig_term(int sig)
{
	_TP_SigTerm = 1;
	signal(sig, _sig_term);
}

static int _socket_set_nonblocking(int fd)
{
	int		res;

#ifdef FIONBIO
	res = 1;
	if (ioctl(fd, FIONBIO, &res) < 0)
		return errno;
#else
    int     nonb = 0;

	nonb = O_NDELAY;
	if ((res = fcntl(fd, F_GETFL, 0)) == -1)
		return errno;
	else if (fcntl(fd, F_SETFL, res | nonb) == -1)
		return errno;
#endif
	return 0;
}

static void _tp_log(char *format, ...)
{
	va_list	ap;

	va_start(ap, format);

	if (_TP_LogSyslog)
	{
		char	buf[512];
		int 	ret = vsprintf(buf, format, ap);

		assert(ret < sizeof(buf));

		if (ret > 0)
			syslog(LOG_NOTICE, "%s", buf);
	}

	vfprintf(stderr, format, ap);
	fputc('\n', stderr);
	fflush(stderr);

	va_end(ap);
}

static char *_tp_next_arg(char **args)
{
	char	*start;
	char	*ptr;

	if (!args || !*args)
		return NULL;
	start = ptr = *args;
	while((*ptr == '\t') || (*ptr == ' '))
		ptr++;
	if (*ptr == '\0')
	{
		*args = NULL;
		return NULL;
	}
	while((*ptr != '\0') && (*ptr != '\t') && (*ptr != ' '))
	{
		ptr++;
	}
	if (*ptr != '\0')
	{
		*ptr++ = '\0';
	}
	*args = ptr;
	return start;
}

static void _tp_free_conn(TCPPROXY_CONN *tpconn)
{
	TCPPROXY_CONFIG		*tpc = tpconn->tpc_config;
	TCPPROXY_PROXY		*tpp = tpconn->tpc_proxy;
	TCPPROXY_POLLTYPE	*pt;

	if (tpconn->tpc_prev)
		tpconn->tpc_prev->tpc_next = tpconn->tpc_next;
	else
		tpp->tpp_conns = tpconn->tpc_next;
	if (tpconn->tpc_next != NULL)
		tpconn->tpc_next->tpc_prev = tpconn->tpc_prev;

	if (tpconn->tpc_afd > 0)
	{
		if (--tpc->tpc_pfds_num != tpconn->tpc_afd_pfds_pos)
		{
			pt = tpc->tpc_conns[
					tpc->tpc_pfds[tpc->tpc_pfds_num].fd];
			assert(pt != NULL);
			if (pt->pt_type == TCPPROXY_PT_TYPE_PROXY)
			{
				((TCPPROXY_PROXY *)pt->pt_vptr)->tpp_lfd_pfds_pos =
					tpconn->tpc_afd_pfds_pos;
			}
			else
			{	
				if (tpc->tpc_pfds[tpc->tpc_pfds_num].fd ==
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd)
				{
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd_pfds_pos =
						tpconn->tpc_afd_pfds_pos;
				}
				else
				{
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_cfd_pfds_pos =
						tpconn->tpc_afd_pfds_pos;
				}
			}
			tpc->tpc_pfds[tpconn->tpc_afd_pfds_pos] =
				tpc->tpc_pfds[tpc->tpc_pfds_num];
		}
		close(tpconn->tpc_afd);
		tpconn->tpc_config->tpc_conns[tpconn->tpc_afd] = NULL;
	}
	if (tpconn->tpc_cfd > 0)
	{
		if (--tpc->tpc_pfds_num != tpconn->tpc_cfd_pfds_pos)
		{
			pt = tpc->tpc_conns[
					tpc->tpc_pfds[tpc->tpc_pfds_num].fd];
			assert(pt != NULL);
			if (pt->pt_type == TCPPROXY_PT_TYPE_PROXY)
			{
				((TCPPROXY_PROXY *)pt->pt_vptr)->tpp_lfd_pfds_pos =
					tpconn->tpc_cfd_pfds_pos;
			}
			else
			{	
				if (tpc->tpc_pfds[tpc->tpc_pfds_num].fd ==
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd)
				{
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd_pfds_pos =
						tpconn->tpc_cfd_pfds_pos;
				}
				else
				{
					((TCPPROXY_CONN *)pt->pt_vptr)->tpc_cfd_pfds_pos =
						tpconn->tpc_cfd_pfds_pos;
				}
			}
			tpc->tpc_pfds[tpconn->tpc_cfd_pfds_pos] =
				tpc->tpc_pfds[tpc->tpc_pfds_num];
		}
		close(tpconn->tpc_cfd);
		tpconn->tpc_config->tpc_conns[tpconn->tpc_cfd] = NULL;
	}

	if (tpconn->tpc_afd_outqueue.q_len)
		_tp_queue_del(&tpconn->tpc_afd_outqueue,
					tpconn->tpc_afd_outqueue.q_len);
	if (tpconn->tpc_cfd_outqueue.q_len)
		_tp_queue_del(&tpconn->tpc_cfd_outqueue,
					tpconn->tpc_cfd_outqueue.q_len);

	free(tpconn);
}

static void _tp_free_conns(TCPPROXY_PROXY *tpp)
{
	while(tpp->tpp_conns)
		_tp_free_conn(tpp->tpp_conns);
}

static void _tp_free_allows(TCPPROXY_ALLOW **allows)
{
	TCPPROXY_ALLOW	*a;
	TCPPROXY_ALLOW	*nexta;

	for(a=*allows;a;a=nexta)
	{
		nexta = a->tpa_next;
		free(a);
		a = nexta;
	}
	*allows = NULL;
}

static int _tp_allowed(TCPPROXY_ALLOW *allows, struct sockaddr_in *sin)
{
	while(allows)
	{
		if (allows->tpa_inaddr.s_addr == sin->sin_addr.s_addr)
			return 1;
		allows = allows->tpa_next;
	}
	return 0;
}

static void _tp_config_free(TCPPROXY_CONFIG **tpcptr)
{
	TCPPROXY_CONFIG		*tpc;
	TCPPROXY_PROXY		*tpp;
	TCPPROXY_PROXY		*nexttpp;
	TCPPROXY_POLLTYPE	*pt;

	for(tpp=(*tpcptr)->tpc_proxies;tpp;tpp=nexttpp)
	{
		nexttpp = tpp->tpp_next;
		tpc = tpp->tpp_config;
		if (tpp->tpp_lfd >= 0)
		{
			close(tpp->tpp_lfd);
			if (--tpc->tpc_pfds_num != tpp->tpp_lfd_pfds_pos)
			{
				pt = tpc->tpc_conns[
					tpc->tpc_pfds[tpc->tpc_pfds_num].fd];
				assert(pt != NULL);
				if (pt->pt_type == TCPPROXY_PT_TYPE_PROXY)
				{
					((TCPPROXY_PROXY *)pt->pt_vptr)->tpp_lfd_pfds_pos =
						tpp->tpp_lfd_pfds_pos;
				}
				else
				{	
					if (tpc->tpc_pfds[tpc->tpc_pfds_num].fd ==
						((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd)
					{
						((TCPPROXY_CONN *)pt->pt_vptr)->tpc_afd_pfds_pos =
							tpp->tpp_lfd_pfds_pos;
					}
					else
					{
						((TCPPROXY_CONN *)pt->pt_vptr)->tpc_cfd_pfds_pos =
							tpp->tpp_lfd_pfds_pos;
					}
				}
				tpc->tpc_pfds[tpp->tpp_lfd_pfds_pos] =
					tpc->tpc_pfds[tpc->tpc_pfds_num];
			}
			tpp->tpp_config->tpc_conns[tpp->tpp_lfd] = NULL;
		}
		_tp_free_conns(tpp);
		_tp_free_allows(&(tpp->tpp_allows));
		free(tpp);
	}
	_tp_free_allows(&((*tpcptr)->tpc_allows));
	free(*tpcptr);
	*tpcptr = NULL;
}

static int __tp_config_load(char *filename, TCPPROXY_CONFIG **tpcptr)
{
	int					fd;
	int					ret;
	struct stat			sb;
	char				*start;
	char				*ending;
	int					lineno;
	char				*data;
	char				*ptr;
	char				*tptr;
	char				*command;
	char				*src_addra;
	char				*dest_addra;
	char				*bind_addra;
	struct sockaddr_in	src_addr;
	struct sockaddr_in	dest_addr;
	struct sockaddr_in	bind_addr;
	TCPPROXY_PROXY		*tpp;
	TCPPROXY_PROXY		*lasttpp = NULL;
	TCPPROXY_CONFIG		*tpc;

	if (strlen(filename) > 128)
		return EINVAL;

	fd = open(filename, O_RDONLY);
	if (fd < 0)
		return errno;
	ret = fstat(fd, &sb);
	if (ret < 0)
	{
		ret = errno;
		close(fd);
		return ret;
	}
	data = (char *)malloc(sb.st_size + 1);
	if (data == NULL)
	{
		ret = errno;
		close(fd);
		return ret;
	}

	ret = read(fd, data, sb.st_size);
	if (ret < 0)
	{
		ret = errno;
		free(data);
		close(fd);
		return ret;
	}
	else if (ret != sb.st_size)
	{
		free(data);
		close(fd);
		return EFAULT;
	}
	close(fd);
	data[sb.st_size] = '\0';

	tpc = (TCPPROXY_CONFIG *)calloc(1, sizeof(TCPPROXY_CONFIG));
	if (tpc == NULL)
	{
		ret = errno;
		free(data);
		return ret;
	}

	lineno = 0;
	ending = data + sb.st_size;
	for(ptr=data;ptr<ending;ptr++)
	{
		start = ptr;
		while((*ptr != '\0') && (*ptr != '\r') && (*ptr != '\n'))
			ptr++;
		if ((*ptr == '\r') && (*(ptr+1) == '\n'))
		{
			*ptr++ = '\0';
		}
		*ptr = '\0';
		lineno++;
		if ((*start == '#') || !*start)
			continue;
		command = _tp_next_arg(&start);
		if (!command)
			continue;
		if (!strcasecmp(command, "allow"))
		{
			TCPPROXY_ALLOW	**allows = lasttpp ?
								&(lasttpp->tpp_allows) : &(tpc->tpc_allows);
			TCPPROXY_ALLOW	*a;
			char			*ipaddr = _tp_next_arg(&start);

			if ((ipaddr == NULL) || (*ipaddr == '\0'))
			{
				_tp_log("%s:%i -- Not enough arguments for \"allow\"",
					filename, lineno);
				continue;
			}

			a = (TCPPROXY_ALLOW *)calloc(1, sizeof(TCPPROXY_ALLOW));
			if (a == NULL)
			{
				_tp_log("%s:%i -- Out of memory", filename, lineno);
				continue;
			}

			a->tpa_inaddr.s_addr = inet_addr(ipaddr);
			if (a->tpa_inaddr.s_addr == -1)
			{
				free(a);
				_tp_log("%s:%i -- Invalid IP for \"allow\"", filename, lineno);
				continue;
			}

			a->tpa_prev = NULL;
			if ((a->tpa_next = *allows) != NULL)
				a->tpa_next->tpa_prev = a;
			*allows = a;

			continue;
		}
		else if (!strcasecmp(command, "queue_limit"))
		{
			char			*ql = _tp_next_arg(&start);
			int				ql_val;

			if ((ql == NULL) || (*ql == '\0'))
			{
				_tp_log("%s:%i -- Not enough arguments for \"queue_limit\"",
					filename, lineno);
				continue;
			}

			ql_val = atoi(ql);
			if (ql_val <= 0)
			{
				_tp_log("%s:%i -- Invalid argument for \"queue_limit\"",
					filename, lineno);
				continue;
			}

			if (lasttpp)
			{
				if (lasttpp->tpp_queue_limit)
				{
					_tp_log("%s:%i -- \"queue_limit\" already specified for this proxy",
						filename, lineno);
					continue;
				}
				lasttpp->tpp_queue_limit = ql_val;
			}
			else
			{
				if (tpc->tpc_queue_limit)
				{
					_tp_log("%s:%i -- global \"queue_limit\" already specified",
						filename, lineno);
					continue;
				}
				tpc->tpc_queue_limit = ql_val;
			}
			continue;
		}
		else if (!strcasecmp(command, "connect_timeout"))
		{
			char			*ct = _tp_next_arg(&start);
			int				ct_val;

			if ((ct == NULL) || (*ct == '\0'))
			{
				_tp_log("%s:%i -- Not enough arguments for \"connect_timeout\"",
					filename, lineno);
				continue;
			}

			ct_val = atoi(ct);
			if (ct_val <= 0)
			{
				_tp_log("%s:%i -- Invalid argument for \"connect_timeout\"",
					filename, lineno);
				continue;
			}

			if (lasttpp)
			{
				if (lasttpp->tpp_connect_timeout)
				{
					_tp_log("%s:%i -- \"connect_timeout\" already specified for this proxy",
						filename, lineno);
					continue;
				}
				lasttpp->tpp_connect_timeout = ct_val;
			}
			else
			{
				if (tpc->tpc_connect_timeout)
				{
					_tp_log("%s:%i -- global \"connect_timeout\" already specified",
						filename, lineno);
					continue;
				}
				tpc->tpc_connect_timeout = ct_val;
			}
			continue;
		}
		else if (strcasecmp(command, "proxy"))
		{
			if (strlen(command) > 100)
				command[100] = '\0';
			if (strlen(start) > 100)
				start[100] = '\0';
			_tp_log("%s:%i -- Unknown command: \"%s%s%s\"",
				filename, lineno,
				command, start?" ":"", start?start:"");
			continue;
		}
		memset(&src_addr, 0, sizeof(src_addr));
		src_addr.sin_family = AF_INET;
		src_addr.sin_port = htons(0);
		src_addr.sin_addr.s_addr = htonl(INADDR_ANY);
		bind_addr = dest_addr = src_addr;

		src_addra = _tp_next_arg(&start);
		if (src_addra == NULL)
		{
			_tp_log("%s:%i -- Not enough arguments to \"proxy\"",
					filename, lineno);
			continue;
		}
		if ((tptr = strchr(src_addra, ':')) != NULL)
		{
			*tptr++ = '\0';
			src_addr.sin_addr.s_addr = inet_addr(src_addra);
			if (src_addr.sin_addr.s_addr == -1)
			{
				if (strlen(src_addra) > 50)
					src_addra[50] = '\0';
				if (strlen(tptr) > 50)
					tptr[50] = '\0';
				_tp_log(
					"%s:%i -- Invalid src IP for src address \"%s:%s\"",
					filename, lineno, src_addra, tptr);
				continue;
			}
			src_addr.sin_port = htons(atoi(tptr));
		}
		else
		{
			if (strlen(src_addra) > 100)
				src_addra[100] = '\0';
			_tp_log("%s:%i -- Missing port for src address \"%s\"",
					filename, lineno, src_addra);
			continue;
		}
		dest_addra = _tp_next_arg(&start);
		if (dest_addra == NULL)
		{
			_tp_log("%s:%i -- Not enough arguments to \"proxy\"",
					filename, lineno);
			continue;
		}
		if ((tptr = strchr(dest_addra, ':')) != NULL)
		{
			*tptr++ = '\0';
			dest_addr.sin_addr.s_addr = inet_addr(dest_addra);
			if (dest_addr.sin_addr.s_addr == -1)
			{
				if (strlen(dest_addra) > 50)
					dest_addra[50] = '\0';
				if (strlen(tptr) > 50)
					tptr[50] = '\0';
				_tp_log(
					"%s:%i -- Invalid dest IP for dest address \"%s:%s\"",
					filename, lineno, dest_addra, tptr);
				continue;
			}
			dest_addr.sin_port = htons(atoi(tptr));
		}
		else
		{
			if (strlen(dest_addra) > 100)
				dest_addra[100] = '\0';
			_tp_log(
				"%s:%i -- Missing port for dest address \"%s\"",
					filename, lineno, dest_addra);
			continue;
		}
		bind_addra = _tp_next_arg(&start);
		if (bind_addra != NULL)
		{
			bind_addr.sin_addr.s_addr = inet_addr(bind_addra);
			if (bind_addr.sin_addr.s_addr == -1)
			{
				if (strlen(bind_addra) > 100)
					bind_addra[100] = '\0';
				_tp_log(
					"%s:%i -- Invalid address for bind address \"%s\"",
					filename, lineno, bind_addra);
				continue;
			}
	
		}
		for(tpp=tpc->tpc_proxies;tpp;tpp=tpp->tpp_next)
		{
			if ((tpp->tpp_src_addr.sin_addr.s_addr ==
							src_addr.sin_addr.s_addr) &&
					(tpp->tpp_src_addr.sin_port ==
							src_addr.sin_port))
				break;
		}
		if (tpp)
		{
			_tp_log("%s:%i Duplicate src address \"%s:%u\"",
				filename, lineno, inet_ntoa(tpp->tpp_src_addr.sin_addr), 
				(u_int)ntohs(tpp->tpp_src_addr.sin_port));
			continue;
		}
		tpp = (TCPPROXY_PROXY *)calloc(1, sizeof(TCPPROXY_PROXY));
		if (!tpp)
		{
			ret = errno;
			_tp_config_free(&tpc);
			free(data);
			return ret;
		}
		tpp->tpp_polltype.pt_vptr = tpp;
		tpp->tpp_polltype.pt_type = TCPPROXY_PT_TYPE_PROXY;
		tpp->tpp_config = tpc;
		tpp->tpp_lfd = -1;
		tpp->tpp_src_addr = src_addr;
		tpp->tpp_dest_addr = dest_addr;
		tpp->tpp_bind_addr = bind_addr;
		tpp->tpp_next = tpc->tpc_proxies;
		tpc->tpc_proxies = tpp;
		lasttpp = tpp;
	}

	free(data);

	if (tpcptr)
		*tpcptr = tpc;
	else
		_tp_config_free(&tpc);

	return 0;
}

static int _tp_config_load(char *filename, TCPPROXY_CONFIG **tpcptr)
{
	int				s;
	int				e;
	int				on;
	TCPPROXY_PROXY	*p_new;
	TCPPROXY_PROXY	*p_old;
	TCPPROXY_PROXY	*p_next;
	TCPPROXY_PROXY	*p_last;
	TCPPROXY_ALLOW	*swap;
	TCPPROXY_CONFIG	*tpc_new;
	struct pollfd	*pfdptr;

	if (!tpcptr)
		return EINVAL;

	s = __tp_config_load(filename, &tpc_new);
	if (s)
		return s;

	if (!*tpcptr)
	{
		*tpcptr = (TCPPROXY_CONFIG *)calloc(1, sizeof(TCPPROXY_CONFIG));
		if (*tpcptr == NULL)
		{
			e = errno;
			_tp_config_free(&tpc_new);
			return e;
		}
	}

	/* look for new listeners, first.. */
	for(p_new=tpc_new->tpc_proxies;p_new;p_new=p_new->tpp_next)
	{
		for(p_old=(*tpcptr)->tpc_proxies;p_old;p_old=p_old->tpp_next)
		{
			if ((p_old->tpp_src_addr.sin_addr.s_addr ==
						p_new->tpp_src_addr.sin_addr.s_addr) &&
					(p_old->tpp_src_addr.sin_port ==
						p_new->tpp_src_addr.sin_port))
				break;
		}
		if (p_old && (p_old->tpp_dest_addr.sin_addr.s_addr ==
					p_new->tpp_dest_addr.sin_addr.s_addr) &&
				(p_old->tpp_dest_addr.sin_port ==
					p_new->tpp_dest_addr.sin_port) &&
				(p_old->tpp_bind_addr.sin_addr.s_addr ==
					p_new->tpp_bind_addr.sin_addr.s_addr) &&
				(p_old->tpp_bind_addr.sin_port ==
					p_new->tpp_bind_addr.sin_port))
		{
			p_old->tpp_queue_limit = p_new->tpp_queue_limit;
			p_old->tpp_connect_timeout = p_new->tpp_connect_timeout;

			/* swap allows */
			swap = p_old->tpp_allows;
			p_old->tpp_allows = p_new->tpp_allows;
			p_new->tpp_allows = swap;
			continue;
		}
		else if (p_old)
		{
			p_old->tpp_dest_addr = p_new->tpp_dest_addr;
			p_old->tpp_bind_addr = p_new->tpp_bind_addr;
			p_old->tpp_queue_limit = p_new->tpp_queue_limit;
			p_old->tpp_connect_timeout = p_new->tpp_connect_timeout;

			/* swap allows */
			swap = p_old->tpp_allows;
			p_old->tpp_allows = p_new->tpp_allows;
			p_new->tpp_allows = swap;

			/* close all conns */
			_tp_free_conns(p_old);

			continue;
		}
		s = socket(AF_INET, SOCK_STREAM, 0);
		if (s < 0)
		{
			e = errno;
			_tp_log(
				"Can't listen on %s:%u, socket() failed: %s",
				inet_ntoa(p_new->tpp_src_addr.sin_addr),
				(u_int)ntohs(p_new->tpp_src_addr.sin_port),
				_my_strerror(e));
			continue;
		}
		e = _socket_set_nonblocking(s);
		if (e)
		{
			close(s);
			_tp_log(
				"Can't listen on %s:%u, set_nonblocking() failed: %s",
				inet_ntoa(p_new->tpp_src_addr.sin_addr),
				(u_int)ntohs(p_new->tpp_src_addr.sin_port),
				_my_strerror(e));
			continue;
		}
		on = 1;
		setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
			(char *)&on, sizeof(on));
		on = 1;
		setsockopt(s, SOL_SOCKET, SO_KEEPALIVE,
			(char *)&on, sizeof(on));

		if (bind(s, (struct sockaddr *)&(p_new->tpp_src_addr),
			sizeof(struct sockaddr_in)) < 0)
		{
			e = errno;
			close(s);
			_tp_log(
				"Can't listen on %s:%u, bind() failed: %s",
				inet_ntoa(p_new->tpp_src_addr.sin_addr),
				(u_int)ntohs(p_new->tpp_src_addr.sin_port),
				_my_strerror(e));
			continue;
		}

		if (listen(s, 1024) < 0)
		{
			e = errno;
			close(s);
			_tp_log(
				"Can't listen on %s:%u, listen() failed: %s",
				inet_ntoa(p_new->tpp_src_addr.sin_addr),
				(u_int)ntohs(p_new->tpp_src_addr.sin_port),
				_my_strerror(e));
			continue;
		}

		p_old = (TCPPROXY_PROXY *)calloc(1, sizeof(TCPPROXY_PROXY));
		if (p_old == NULL)
		{
			e = errno;
			close(s);
			_tp_log(
				"Can't listen on %s:%u, calloc failed: %s",
				inet_ntoa(p_new->tpp_src_addr.sin_addr),
				(u_int)ntohs(p_new->tpp_src_addr.sin_port),
				_my_strerror(e));
			continue;
		}

		p_old->tpp_queue_limit = p_new->tpp_queue_limit;
		p_old->tpp_connect_timeout = p_new->tpp_connect_timeout;

		if ((*tpcptr)->tpc_pfds_num == (*tpcptr)->tpc_pfds_size)
		{
			void *vptr = realloc((*tpcptr)->tpc_pfds,
				sizeof(struct pollfd)*((*tpcptr)->tpc_pfds_size+16));
			if (!vptr)
			{
				e = errno;
				close(s);
				free(p_old);
				_tp_log(
					"Can't listen on %s:%u, realloc failed: %s",
					inet_ntoa(p_new->tpp_src_addr.sin_addr),
					(u_int)ntohs(p_new->tpp_src_addr.sin_port),
					_my_strerror(e));
				continue;
			}
			(*tpcptr)->tpc_pfds = (struct pollfd *)vptr;
			(*tpcptr)->tpc_pfds_size += 16;
		}
		if (s >= (*tpcptr)->tpc_conns_size)
		{
			void *vptr;
			vptr = realloc((*tpcptr)->tpc_conns,
				sizeof(TCPPROXY_POLLTYPE *) * (s + 16));
			if (vptr == NULL)
			{
				e = errno;
				close(s);
				free(p_old);
				_tp_log(
					"Can't listen on %s:%u, realloc failed: %s",
					inet_ntoa(p_new->tpp_src_addr.sin_addr),
					(u_int)ntohs(p_new->tpp_src_addr.sin_port),
					_my_strerror(e));
				continue;
			}
			(*tpcptr)->tpc_conns_size = s + 16;
			(*tpcptr)->tpc_conns = (TCPPROXY_POLLTYPE **)vptr;
		}
		p_old->tpp_polltype.pt_vptr = p_old;
		p_old->tpp_polltype.pt_type = TCPPROXY_PT_TYPE_PROXY;
		(*tpcptr)->tpc_conns[s] = &(p_old->tpp_polltype);
		p_old->tpp_config = *tpcptr;
		p_old->tpp_lfd = s;
		p_old->tpp_src_addr = p_new->tpp_src_addr;
		p_old->tpp_dest_addr = p_new->tpp_dest_addr;
		p_old->tpp_bind_addr = p_new->tpp_bind_addr;
		p_old->tpp_next = (*tpcptr)->tpc_proxies;
		(*tpcptr)->tpc_proxies = p_old;
		p_old->tpp_lfd_pfds_pos = (*tpcptr)->tpc_pfds_num++;
		pfdptr = (*tpcptr)->tpc_pfds + p_old->tpp_lfd_pfds_pos;
		pfdptr->fd = s;
		pfdptr->events = POLLIN;
		pfdptr->revents = 0;

		/* swap allows */
		swap = p_old->tpp_allows;
		p_old->tpp_allows = p_new->tpp_allows;
		p_new->tpp_allows = swap;
	}

	/* look for listeners to delete now.. */
	p_last = NULL;
	for(p_old=(*tpcptr)->tpc_proxies;p_old;p_old=p_next)
	{
		p_next = p_old->tpp_next;

		for(p_new=tpc_new->tpc_proxies;p_new;p_new=p_new->tpp_next)
		{
			if ((p_old->tpp_src_addr.sin_addr.s_addr ==
						p_new->tpp_src_addr.sin_addr.s_addr) &&
					(p_old->tpp_src_addr.sin_port ==
						p_new->tpp_src_addr.sin_port))
				break;
		}
		if (p_new)
		{
			p_last = p_old;
			continue;
		}
		if (p_last)
			p_last->tpp_next = p_next;
		else
			(*tpcptr)->tpc_proxies = p_next;
		p_old->tpp_next = tpc_new->tpc_proxies;
		tpc_new->tpc_proxies = p_old;
	}

	/* swap allows */
	swap = tpc_new->tpc_allows;
	tpc_new->tpc_allows = (*tpcptr)->tpc_allows;
	(*tpcptr)->tpc_allows = swap;

	if (!((*tpcptr)->tpc_connect_timeout = tpc_new->tpc_connect_timeout))
		(*tpcptr)->tpc_connect_timeout = CSTP_DEFAULT_CONNECT_TIMEOUT;

	if (!((*tpcptr)->tpc_queue_limit = tpc_new->tpc_queue_limit))
		(*tpcptr)->tpc_queue_limit = CSTP_DEFAULT_QUEUE_LIMIT;

	_tp_config_free(&tpc_new);

	return 0;
}

static int _tp_queue_fill_iovec(TCPPROXY_Q *queue, struct iovec *iov, int num)
{
	TCPPROXY_QE	*qe = queue->q_entries;
	int			num_left = num;

	assert(qe != NULL);

	for(;;)
	{
		assert(!qe->qe_next || (qe->qe_len == qe->qe_size));

		if (qe->qe_len - qe->qe_pos)
		{
			iov->iov_base = qe->qe_data + qe->qe_pos;
			iov->iov_len = qe->qe_len - qe->qe_pos;
			iov++;
			--num_left;
		}
		if (((qe = qe->qe_next) == NULL) || (num_left == 0))
			break;
	}
	return num - num_left;
}

static int _tp_queue_del(TCPPROXY_Q *queue, int amt)
{
	TCPPROXY_QE	*nextqe;
	int			l;

	assert(queue->q_entries != NULL);
	assert(queue->q_len >= amt);

	queue->q_len -= amt;

	while(amt > (l = (queue->q_entries->qe_len - queue->q_entries->qe_pos)))
	{
		nextqe = queue->q_entries->qe_next;
		assert(nextqe);

		free(queue->q_entries);
		amt -= l;
		queue->q_entries = nextqe;
	}

	if (!queue->q_len)
	{
		assert(queue->q_entries->qe_next == NULL);
		queue->q_entries->qe_pos = queue->q_entries->qe_len = 0;
	}
	else
		queue->q_entries->qe_pos += amt;

	assert(queue->q_entries->qe_pos <= queue->q_entries->qe_len);

	return 0;
}

static int _tp_queue_add(TCPPROXY_Q *queue, char *data, int datalen)
{
	TCPPROXY_QE		*qe = queue->q_last_entry;
	int				qe_room;

	queue->q_len += datalen;
	for(;;)
	{
		if ((qe_room = (qe->qe_size - qe->qe_len)) < datalen)
		{
			static int thesize = (sizeof(TCPPROXY_QE)+8192) & ~8191;

			if (qe_room)
			{
				memcpy(qe->qe_data+qe->qe_len, data, qe_room);
				data += qe_room;
				datalen -= qe_room;
				qe->qe_len = qe->qe_size;
			}

			qe  = (TCPPROXY_QE *)malloc(thesize);
			if (!qe)
				return ENOMEM;
			qe->qe_pos = 0;
			qe->qe_size = thesize - sizeof(TCPPROXY_QE);
			qe->qe_data = (char *)(qe+1);
			qe->qe_next = NULL;
			if (queue->q_last_entry)
				queue->q_last_entry->qe_next = qe;
			else
				queue->q_entries = qe;
			queue->q_last_entry = qe;

			if (datalen > qe->qe_size)
			{
				memcpy(qe->qe_data, data, qe->qe_size);
				data += qe->qe_size;
				datalen -= qe->qe_size;
				qe->qe_len = qe->qe_size;

				qe  = (TCPPROXY_QE *)malloc(thesize);
				if (!qe)
					return ENOMEM;
				qe->qe_pos = 0;
				qe->qe_size = thesize - sizeof(TCPPROXY_QE);
				qe->qe_data = (char *)(qe+1);
				qe->qe_len = 0;
				qe->qe_next = NULL;
				if (queue->q_last_entry)
					queue->q_last_entry->qe_next = qe;
				else
					queue->q_entries = qe;
				queue->q_last_entry = qe;
				continue;
			}
			qe->qe_len = 0;
		}
		memcpy(qe->qe_data + qe->qe_len, data, datalen);
		qe->qe_len += datalen;
		break;
	}
	return 0;
}

static int _tp_add_conn(TCPPROXY_PROXY *tpp, int fd, struct sockaddr_in *sin)
{
	static int		thesize = (sizeof(TCPPROXY_QE)+8192) & ~8191;
	TCPPROXY_CONN	*tpc;
	TCPPROXY_CONFIG	*tpcfg;
	int				ret;
	int				on;
	struct pollfd	*pfdptr;
	int				high_fd;
	int				e;
	TCPPROXY_QE		*qe;
	struct timeval	tv;

	on = 1;
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
	on = 1;
	setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));
	on = 1024*64;
	if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&on, sizeof(on)))
	{
		on = 1024*32;
		setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&on, sizeof(on));
	}
	on = 1024*64;
	if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&on, sizeof(on)))
	{
		on = 1024*32;
		setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&on, sizeof(on));
	}
	tpc = (TCPPROXY_CONN *)calloc(1,sizeof(TCPPROXY_CONN));
	if (tpc == NULL)
	{
		e = errno;
		_tp_log(
			"Can't connect to %s:%u, socket() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	gettimeofday(&tv, NULL);
	tpc->tpc_conntime = tv.tv_sec;

	qe  = (TCPPROXY_QE *)malloc(thesize);
	if (!qe)
	{
		e = errno;
		free(tpc);
		_tp_log(
			"Can't connect to %s:%u, socket() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	qe->qe_len = 0;
	qe->qe_pos = 0;
	qe->qe_size = thesize - sizeof(TCPPROXY_QE);
	qe->qe_data = (char *)(qe+1);
	qe->qe_next = NULL;
	tpc->tpc_afd_outqueue.q_entries = qe;
	tpc->tpc_afd_outqueue.q_last_entry = qe;
	qe  = (TCPPROXY_QE *)malloc(thesize);
	if (!qe)
	{
		e = errno;
		free(tpc);
		_tp_log(
			"Can't connect to %s:%u, socket() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	qe->qe_len = 0;
	qe->qe_pos = 0;
	qe->qe_size = thesize - sizeof(TCPPROXY_QE);
	qe->qe_data = (char *)(qe+1);
	qe->qe_next = NULL;
	tpc->tpc_cfd_outqueue.q_entries = qe;
	tpc->tpc_cfd_outqueue.q_last_entry = qe;
	tpc->tpc_polltype.pt_vptr = tpc;
	tpc->tpc_polltype.pt_type = TCPPROXY_PT_TYPE_CONN;
	tpc->tpc_afd = fd;
	tpc->tpc_rem_addr = *sin;
	tpc->tpc_proxy = tpp;
	tpcfg = tpc->tpc_config = tpp->tpp_config;
	tpc->tpc_cfd = socket(AF_INET, SOCK_STREAM, 0);
	if (tpc->tpc_cfd < 0)
	{
		e = errno;
		free(tpc);
		_tp_log(
			"Can't connect to %s:%u, socket() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	e = _socket_set_nonblocking(tpc->tpc_cfd);
	if (e)
	{
		close(tpc->tpc_cfd);
		free(tpc);
		_tp_log(
			"Can't connect to %s:%u, set_nonblocking() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	on = 1;
	setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_REUSEADDR,
		(char *)&on, sizeof(on));
	on = 1;
	setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_KEEPALIVE,
		(char *)&on, sizeof(on));
	on = 1024*64;
	if (setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_RCVBUF, (char *)&on, sizeof(on)))
	{
		on = 1024*32;
		setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_RCVBUF, (char *)&on, sizeof(on));
	}
	on = 1024*64;
	if (setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_SNDBUF, (char *)&on, sizeof(on)))
	{
		on = 1024*32;
		setsockopt(tpc->tpc_cfd, SOL_SOCKET, SO_SNDBUF, (char *)&on, sizeof(on));
	}
	ret = bind(tpc->tpc_cfd, (struct sockaddr *)&(tpp->tpp_bind_addr),
				sizeof(tpp->tpp_bind_addr));
	if (ret < 0)
	{
		e = errno;
		close(tpc->tpc_cfd);
		free(tpc);
		_tp_log(
			"Can't connect to %s:%u, bind() failed: %s",
			inet_ntoa(tpp->tpp_dest_addr.sin_addr),
			(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
			_my_strerror(e));
		return e;
	}
	ret = connect(tpc->tpc_cfd, (struct sockaddr *)&(tpp->tpp_dest_addr),
				sizeof(tpp->tpp_dest_addr));
	if (ret < 0)
	{
		if (errno != EINPROGRESS)
		{
			e = errno;
			close(tpc->tpc_cfd);
			free(tpc);
			_tp_log(
				"Can't connect to %s:%u, connect() failed: %s",
				inet_ntoa(tpp->tpp_dest_addr.sin_addr),
				(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
				_my_strerror(e));
			return e;
		}
	}
	if (tpcfg->tpc_pfds_num+1 >= tpcfg->tpc_pfds_size)
	{
		void *vptr = realloc(tpcfg->tpc_pfds,
			sizeof(struct pollfd)*(tpcfg->tpc_pfds_size+16));
		if (vptr == NULL)
		{
			e = errno;
			close(tpc->tpc_cfd);
			free(tpc);
			_tp_log(
				"Can't connect to %s:%u, realloc failed: %s",
				inet_ntoa(tpp->tpp_dest_addr.sin_addr),
				(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
				_my_strerror(e));
			return e;
		}
		tpcfg->tpc_pfds = (struct pollfd *)vptr;
		tpcfg->tpc_pfds_size += 16;
	}

	high_fd = tpc->tpc_cfd > tpc->tpc_afd ? tpc->tpc_cfd : tpc->tpc_afd;

	if (high_fd >= tpcfg->tpc_conns_size)
	{
		void *vptr;
		vptr = realloc(tpcfg->tpc_conns,
				sizeof(TCPPROXY_POLLTYPE *) * (high_fd + 16));
		if (vptr == NULL)
		{
			e = errno;
			close(tpc->tpc_cfd);
			free(tpc);
			_tp_log(
				"Can't connect to %s:%u, realloc failed: %s",
				inet_ntoa(tpp->tpp_dest_addr.sin_addr),
				(u_int)ntohs(tpp->tpp_dest_addr.sin_port),
				_my_strerror(e));
			return e;
		}
		tpcfg->tpc_conns_size = high_fd + 16;
		tpcfg->tpc_conns = (TCPPROXY_POLLTYPE **)vptr;
	}

	tpcfg->tpc_conns[tpc->tpc_cfd] = &tpc->tpc_polltype;
	tpcfg->tpc_conns[tpc->tpc_afd] = &tpc->tpc_polltype;
	tpc->tpc_cfd_pfds_pos = tpcfg->tpc_pfds_num++;
	pfdptr = tpcfg->tpc_pfds + tpc->tpc_cfd_pfds_pos;
	pfdptr->fd = tpc->tpc_cfd;
	pfdptr->events = POLLOUT;
	pfdptr->revents = 0;
	tpc->tpc_afd_pfds_pos = tpcfg->tpc_pfds_num++;
	pfdptr = tpcfg->tpc_pfds + tpc->tpc_afd_pfds_pos;
	pfdptr->fd = tpc->tpc_afd;
	pfdptr->events = POLLIN;
	pfdptr->revents = 0;
	tpc->tpc_flags = TCPPROXY_TPC_FLAGS_CONNECTING;

	tpc->tpc_prev = NULL;
	if ((tpc->tpc_next = tpp->tpp_conns) != NULL)
		tpc->tpc_next->tpc_prev = tpc;
	tpp->tpp_conns = tpc;

	return 0;
}

static int _tp_conn_complete(TCPPROXY_CONN *tpc, struct pollfd *pfd)
{
	int ret = connect(tpc->tpc_cfd,
			(struct sockaddr *)&(tpc->tpc_proxy->tpp_dest_addr),
			sizeof(tpc->tpc_proxy->tpp_dest_addr));

	if ((ret < 0) && (errno != EISCONN))
		return errno;
	tpc->tpc_flags &= ~TCPPROXY_TPC_FLAGS_CONNECTING;
	if (!tpc->tpc_cfd_outqueue.q_len)
		pfd->events &= ~POLLOUT;
	pfd->events |= POLLIN;
	return 0;
}

static void _tp_ioloop(TCPPROXY_CONFIG *tpc)
{
	TCPPROXY_CONN		*tpconn;
	TCPPROXY_POLLTYPE	*pt;
	struct pollfd		*pfdptr;
	int					ret;
	int					nfds;
	int					i;
	TCPPROXY_Q			*q;
	int					pos;
	int					re;
	char				read_buffer[128*1024];
	char				*wptr;
	struct timeval		tv;

	gettimeofday(&tv, NULL);

	if (!tpc->tpc_timeout_check)
		tpc->tpc_timeout_check = tv.tv_sec + 10;

	for(;;)
	{
		if (_TP_SigHup || _TP_SigTerm)
			break;

		if (!tpc->tpc_pfds_num)
		{
			sleep(1);
			continue;
		}

		nfds = poll(tpc->tpc_pfds, tpc->tpc_pfds_num, 5000);
		if (nfds < 0)
			continue;
		gettimeofday(&tv, NULL);
		for(i=0;i<tpc->tpc_pfds_num;++i)
		{
			if (!nfds && (tv.tv_sec < tpc->tpc_timeout_check))
				break;
			if ((re = tpc->tpc_pfds[i].revents) != 0)
			{
				--nfds;
				tpc->tpc_pfds[i].revents = 0;
			}
			pfdptr = tpc->tpc_pfds + i;

			assert(pfdptr->fd < tpc->tpc_conns_size);
			pt = tpc->tpc_conns[pfdptr->fd];
			if (pt == NULL)
				continue;

			assert(pt->pt_vptr != NULL);
			if (pt->pt_type == TCPPROXY_PT_TYPE_PROXY)
			{
				/* listener */
				TCPPROXY_PROXY		*tpp = (TCPPROXY_PROXY *)pt->pt_vptr;
				struct sockaddr_in	rem_sin;
				int					nsock;
#ifdef _ALL_SOURCE
				long				len;
#else
				int					len;
#endif
				int					ret;

				if (!re)
					continue;

				len = sizeof(rem_sin);
				nsock = accept(tpp->tpp_lfd,
					(struct sockaddr *)&rem_sin, &len);
				if (nsock < 0)
					continue;
				else if (len != sizeof(rem_sin))
				{
					close(nsock);
					continue;
				}

				if (!_tp_allowed(tpp->tpp_allows, &rem_sin) &&
					!_tp_allowed(tpc->tpc_allows, &rem_sin))
				{
					close(nsock);
					continue;
				}

				ret = _socket_set_nonblocking(nsock);
				if (ret)
				{
					close(nsock);
					_tp_log("Couldn't set_nonblocking() for new connection from %s: %s",
						inet_ntoa(rem_sin.sin_addr),
						_my_strerror(ret));
					continue;
				}

				/* add new conn */
				if ((ret = _tp_add_conn(tpp, nsock, &rem_sin)))
				{
					close(nsock);
					_tp_log("add_conn() failed for new connection from %s: %s",
						inet_ntoa(rem_sin.sin_addr),
						_my_strerror(ret));
				}

				continue;
			}

			assert(pt->pt_type == TCPPROXY_PT_TYPE_CONN);

			tpconn = (TCPPROXY_CONN *)pt->pt_vptr;

			if ((tpconn->tpc_flags & TCPPROXY_TPC_FLAGS_CONNECTING) &&
				(tv.tv_sec > (tpconn->tpc_conntime +
						(tpconn->tpc_proxy->tpp_connect_timeout ? 
							tpconn->tpc_proxy->tpp_connect_timeout :
							tpconn->tpc_config->tpc_connect_timeout))))
			{
				_tp_free_conn(tpconn);
				continue;
			}

			if (!re)
				continue;

			if (re & POLLOUT)
			{
#ifdef IOV_MAX
#define NUM_IOV	IOV_MAX
#else
#define NUM_IOV		8
#endif
				struct iovec	iov[NUM_IOV];
				int				num;

				if (tpconn->tpc_flags & TCPPROXY_TPC_FLAGS_CONNECTING)
				{
					assert(pfdptr->fd == tpconn->tpc_cfd);

					ret = _tp_conn_complete(tpconn, pfdptr);
					if (ret)
					{
						_tp_free_conn(tpconn);
						continue;
					}
					continue;
				}

				q = (pfdptr->fd == tpconn->tpc_afd) ?
						&tpconn->tpc_afd_outqueue :
						&tpconn->tpc_cfd_outqueue;

				num = _tp_queue_fill_iovec(q, iov, NUM_IOV);
				if (num < 0)
				{
					_tp_free_conn(tpconn);
					continue;
				}

				if (!num)
				{
					pfdptr->events &= ~POLLOUT;
				}
				else
				{
					ret = writev(pfdptr->fd, iov, num);
					if (ret > 0)
					{
						ret = _tp_queue_del(q, ret);
						if (ret)
						{
							_tp_free_conn(tpconn);
							continue;
						}
						if (!q->q_len)
							pfdptr->events &= ~POLLOUT;
					}
				}
				re &= ~POLLOUT;
			}

			if (!re)
				continue;

			ret = read(pfdptr->fd, read_buffer, sizeof(read_buffer));
			if (!ret)
			{
				_tp_free_conn(tpconn);
				continue;
			}
			else if (ret < 0)
			{
				if ((errno == EAGAIN) || (errno == EINTR)
#ifdef EWOULDBLOCK
					|| (errno == EWOULDBLOCK)
#endif
					)
				{
					continue;
				}
				_tp_free_conn(tpconn);
				continue;
			}
			if (pfdptr->fd == tpconn->tpc_afd)
			{
				q = &tpconn->tpc_cfd_outqueue;
				pos = tpconn->tpc_cfd_pfds_pos;
			}
			else
			{
				q = &tpconn->tpc_afd_outqueue;
				pos = tpconn->tpc_afd_pfds_pos;
			}
			wptr = read_buffer;
			if (!q->q_len)
			{
				int nw = write(tpc->tpc_pfds[pos].fd, wptr, ret);
				if (nw > 0)
				{
					wptr += nw;
					ret -= nw;
				}
			}
			if (ret)
			{
				ret = _tp_queue_add(q, wptr, ret);
				if (ret)
				{
					_tp_free_conn(tpconn);
					continue;
				}
				if (q->q_len >
						(tpconn->tpc_proxy->tpp_queue_limit ?
							tpconn->tpc_proxy->tpp_queue_limit :
							tpc->tpc_queue_limit))
				{
					_tp_free_conn(tpconn);
					continue;
				}
				if (!(tpconn->tpc_flags & TCPPROXY_TPC_FLAGS_CONNECTING))
					tpc->tpc_pfds[pos].events |= POLLOUT;
			}
		}
		if (tv.tv_sec >= tpc->tpc_timeout_check)
		{
			tpc->tpc_timeout_check = tv.tv_sec + 10;
		}
	}
}

static void _usage(int l)
{
	fprintf(stderr, "Usage: cstcpproxy [-h] [-n] [-s] config.file\n");
	if (l)
	{
		fprintf(stderr, "    -h  -- this usage\n");
		fprintf(stderr, "    -n  -- don't fork into background\n");
		fprintf(stderr, "    -s  -- log to syslog as well as stderr\n");
	}
}

int main(int argc, char **argv)
{
	TCPPROXY_CONFIG		*TPConfig = NULL;
	int					ret;
	int					dontfork = 0;

	printf("cstcpproxy Version %s by Chris Behrens (cbehrens@codestud.com)\n", CSTP_VERSION);

	while((ret = getopt(argc, argv, "nsh")) != EOF)
	{
		switch(ret)
		{
			case 's':
				_TP_LogSyslog = 1;
				break;
			case 'n':
				dontfork = 1;
				break;
			case 'h':
				_usage(1);
				exit(0);
			default:
				_usage(0);
				exit(-1);
		}
	}

	argc -= optind;
	argv += optind;

	if (argc != 1)
	{
		_usage(0);
		exit(-1);
	}

	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, _sig_hup);
	signal(SIGTERM, _sig_term);

	ret = _tp_config_load(argv[0], &TPConfig);
	if (ret)
	{
		fprintf(stderr, "Load of \"%s\" failed: %s\n",
				argv[1], _my_strerror(ret));
		exit(-1);
	}

	if (!dontfork)
	{
		if ((ret = fork()))
		{
			if (ret == -1)
			{
				fprintf(stderr, "Couldn't fork into the background\n");
			}
			else
			{
				printf("Now running in the background\n");
				exit(0);
			}
		}
	}

	for(;;)
	{
		_tp_ioloop(TPConfig);
		if (_TP_SigTerm)
			break;
		if (_TP_SigHup)
		{
			_TP_SigHup = 0;
			_tp_config_load(argv[0], &TPConfig);
		}
	}

	_tp_config_free(&TPConfig);

	exit(0);
}
